package ttt.gui;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;
import java.util.List;

import ttt.model.Player;
import ttt.model.TTTGame;

/**
 * A view of the Tic-Tac-Toe game that has animation and images to display 
 * the board.
 * 
 * (The animation code is tricky -- I don't expect you to understand all of it 
 * for this course.)
 */
public class HomestarView extends View {
    // class constants
    private static final long serialVersionUID = 0;
    private static final String EXTENSION = ".gif";
    private static final int DELAY = 20;
    
    // fields
    private int width, height, squareWidth, squareHeight;
    private List<Spinner> spinners;
    
    /** Constructs a view object that shows the given game. */
    public HomestarView(TTTGame game) {
        super(game);
        setBackground(Color.magenta.darker());
        
        game.addObserver(this);
        
        addMouseListener(new TTTMouseAdapter());
        spinners = new ArrayList<Spinner>();
    }
    
    /** Paints this view on the screen. */
    public void paintComponent(Graphics g) {
        super.paintComponent(g);
        width = getWidth();
        height = getHeight();
        squareWidth = getSquareWidth();
        squareHeight = getSquareHeight();
        
        Graphics2D g2 = (Graphics2D) g;
        drawGrid(g2, Color.yellow, new BasicStroke(2f));
        
        // draw each square [row, col]
        for (int row = 0; row < TTTGame.BOARD_SIZE; row++) {
            for (int col = 0; col < TTTGame.BOARD_SIZE; col++) {
                drawSquare(g2, row, col);
            }
        }
    }
    
    // Loads the image from the given file.
    private Image createImage(String filename) {
        Image image = null;
        Toolkit tk = Toolkit.getDefaultToolkit();
        image = tk.getImage(filename);
        
        // wait for image to load
        MediaTracker mt = new MediaTracker(this);
        mt.addImage(image, 0);
        try {
            mt.waitForAll();
        } catch (InterruptedException ie) {
        }
        
        return image;
    }
    
    // Draws the grid lines on the panel.
    private void drawGrid(Graphics2D g2, Color c, Stroke s) {
        g2.setPaint(c);
        g2.setStroke(s);
        
        // horizontal line
        for (int i = 1; i < TTTGame.BOARD_SIZE; i++)
            g2.draw(new Line2D.Double(0, i * squareHeight, width, i
                    * squareHeight));
        
        // vertical lines
        for (int i = 1; i < TTTGame.BOARD_SIZE; i++)
            g2.draw(new Line2D.Double(i * squareWidth, 0, i * squareWidth,
                    height));
    }
    
    // Drawing helper that draws the given square.
    // Possibly have to deal with animated "spinners" on this square.
    private void drawSquare(Graphics2D g2, int row, int col) {
        Player square = game.getSquare(row, col);
        if (square == TTTGame.EMPTY)
            return;
        
        int xx = col * squareWidth;
        int yy = row * squareHeight;
        int imageWidth = squareWidth;
        int imageHeight = squareHeight;
        int imagex = xx;
        int imagey = yy;
        
        int spinx = xx + squareWidth / 2;
        int spiny = yy + squareHeight / 2;
        Composite old = g2.getComposite();
        
        // (possibly) rotate this square
        Spinner spin = null;
        for (int i = 0; i < spinners.size(); i++) {
            Spinner next = (Spinner) spinners.get(i);
            
            if (next.row == row && next.col == col && next.angle != 0) {
                spin = next;
                g2.rotate(spin.angle, spinx, spiny);
                imageWidth = (int) (spin.scaleFactor * squareWidth);
                imageHeight = (int) (spin.scaleFactor * squareHeight);
                imagex = xx - (imageWidth - squareWidth) / 2;
                imagey = yy - (imageHeight - squareHeight) / 2;
                
                g2.setComposite(AlphaComposite.getInstance(
                        AlphaComposite.SRC_OVER, Math.min(1.0f, spin.alpha)));
                
                break;
            }
        }
        
        // draw image for this square
        Image img = createImage(IMAGE_FOLDER + square + EXTENSION);
        g2.drawImage(img, imagex, imagey, imageWidth,
                     imageHeight, this);
        
        // set back to normal after possible rotation
        if (spin != null) {
            g2.rotate(-spin.angle, spinx, spiny);
            g2.setComposite(old);
        }
    }
    
    private int getSquareHeight() {
        return getHeight() / TTTGame.BOARD_SIZE;
    }
    
    private int getSquareWidth() {
        return getWidth() / TTTGame.BOARD_SIZE;
    }
    
    /** Handles mouse clicks on this view panel. */
    private class TTTMouseAdapter extends MouseAdapter {
        public void mousePressed(MouseEvent event) {
            int xx = event.getX(), yy = event.getY();
            int col = xx / squareWidth, row = yy / squareHeight;
            
            if (game.canPlay(row, col)) {
                game.play(row, col);
                new Spinner(row, col);   // start animated "spinner"
            }
        }
    }
    
    // an object for a rotating play on the board
    private class Spinner implements ActionListener {
        // constants
        private static final double ANGLE_INCR = Math.PI / 50;
        private static final int NUM_TICKS = (int) (2 * Math.PI / ANGLE_INCR);
        private static final double SCALE_DECR = (3.0 - 1.0) / NUM_TICKS;
        private static final float ALPHA_INCR = 1.0f / NUM_TICKS;
        
        // fields
        private int row = 0;
        private int col = 0;
        private double scaleFactor = 3.0;
        private double angle = 0.1;
        private float alpha = 0.0f;
        
        private javax.swing.Timer timer = new javax.swing.Timer(DELAY, this);
        
        public Spinner(int row, int col) {
            this.row = row;
            this.col = col;
            
            spinners.add(this);
            this.timer.start();
        }
        
        /** Called when animation timer ticks, to update the spinning image */
        public void actionPerformed(ActionEvent event) {
            this.scaleFactor -= SCALE_DECR;
            this.alpha += ALPHA_INCR;
            this.angle += ANGLE_INCR;
            
            if (this.angle > 2 * Math.PI) {
                this.angle = 0;
                this.timer.stop();
                spinners.remove(this);
            }
            
            repaint();
        }
    }
}
